
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>

#include "interface.h"
#include "filesystem.h"
#include "eiface.h"
#include "cbase.h"
#include "vstdlib/random.h"
#include "gmodinterface/gmluamodule.h"
#include "ivmodelinfo.h"
#include "materialsystem/imaterialsystem.h"
#include "materialsystem/itexture.h"
#include "materialsystem/imaterial.h"
#include "icliententitylist.h"
#include "mathlib/vplane.h"

#include "lua_utility.h"

#include "macro.h"

#include "tier0/memdbgon.h"

// interfaces we require for this module
IMaterialSystem* materialsystem = NULL;
IClientEntityList* cliententitylist = NULL;

// setup module
GMOD_MODULE( Init, Shutdown );


// structure for accessing the clip planes on the entity
#pragma pack( push )
#pragma pack( 1 )
struct ClipPlane
{

	float plane[ 4 ];
	bool enabled;

};
#pragma pack( pop )

// offsets
enum Offsets
{

	OFFSET_CLIPPLANE = 1144,
	OFFSET_FOGPARAMS = 5400,
	OFFSET_LOCALDATA = 5476,

};

// mesh stuff
#define TYPE_MESHBUILDER 1900
#define TYPE_MESH 1901

/*-------------------------------------
	PushCustomClipPlane
-------------------------------------*/
LUA_FUNCTION( PushCustomClipPlane )
{

	// plane to clip
	float clipplane[ 4 ];

	// check type
	switch( Lua()->GetType( 1 ) )
	{

	// vector and distance
	case GLua::TYPE_VECTOR:
		{

			// make sure they passed a distance
			Lua()->CheckType( 2, GLua::TYPE_NUMBER );

			// get vector
			Vector& normal = GetVector( L, 1 );
			float dist = Lua()->GetNumber( 2 );

			// set
			clipplane[ 0 ] = normal.x;
			clipplane[ 1 ] = normal.y;
			clipplane[ 2 ] = normal.z;
			clipplane[ 3 ] = dist;

			break;

		}

	}

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->PushCustomClipPlane( clipplane );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	PopCustomClipPlane
-------------------------------------*/
LUA_FUNCTION( PopCustomClipPlane )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->PopCustomClipPlane();
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	EnableClipping
-------------------------------------*/
LUA_FUNCTION( EnableClipping )
{

	Lua()->CheckType( 1, GLua::TYPE_BOOL );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->EnableClipping( Lua()->GetBool( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilEnable
-------------------------------------*/
LUA_FUNCTION( SetStencilEnable )
{

	Lua()->CheckType( 1, GLua::TYPE_BOOL );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilEnable( Lua()->GetBool( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilFailOperation
-------------------------------------*/
LUA_FUNCTION( SetStencilFailOperation )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilFailOperation( (StencilOperation_t)Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilZFailOperation
-------------------------------------*/
LUA_FUNCTION( SetStencilZFailOperation )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilZFailOperation( (StencilOperation_t)Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilPassOperation
-------------------------------------*/
LUA_FUNCTION( SetStencilPassOperation )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilPassOperation( (StencilOperation_t)Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilCompareFunction
-------------------------------------*/
LUA_FUNCTION( SetStencilCompareFunction )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilCompareFunction( (StencilComparisonFunction_t)Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilReferenceValue
-------------------------------------*/
LUA_FUNCTION( SetStencilReferenceValue )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilReferenceValue( Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilTestMask
-------------------------------------*/
LUA_FUNCTION( SetStencilTestMask )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilTestMask( Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetStencilWriteMask
-------------------------------------*/
LUA_FUNCTION( SetStencilWriteMask )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetStencilWriteMask( Lua()->GetInteger( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	ClearStencilBufferRectangle
-------------------------------------*/
LUA_FUNCTION( ClearStencilBufferRectangle )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );
	Lua()->CheckType( 5, GLua::TYPE_NUMBER );

	// arguments
	int x = Lua()->GetInteger( 1 );
	int y = Lua()->GetInteger( 2 );
	int xmax = Lua()->GetInteger( 3 );
	int ymax = Lua()->GetInteger( 4 );
	int value = Lua()->GetInteger( 5 );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->ClearStencilBufferRectangle( x, y, xmax, ymax, value );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	CreateRenderTargetTexture
-------------------------------------*/
LUA_FUNCTION( CreateRenderTargetTexture )
{

	Lua()->CheckType( 1, GLua::TYPE_STRING );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );
	Lua()->CheckType( 5, GLua::TYPE_NUMBER );
	Lua()->CheckType( 6, GLua::TYPE_NUMBER );

	// arguments
	const char* name = Lua()->GetString( 1 );
	int w = Lua()->GetInteger( 2 );
	int h = Lua()->GetInteger( 3 );
	RenderTargetSizeMode_t sizemode = static_cast<RenderTargetSizeMode_t>( Lua()->GetInteger( 4 ) );
	ImageFormat format = static_cast<ImageFormat>( Lua()->GetInteger( 5 ) );
	MaterialRenderTargetDepth_t depth = static_cast<MaterialRenderTargetDepth_t>( Lua()->GetInteger( 6 ) );

	// push or create
	ITexture* texture = materialsystem->FindTexture( name, TEXTURE_GROUP_RENDER_TARGET, false );
	if( !texture || texture->IsError() )
	{

		texture = materialsystem->CreateNamedRenderTargetTexture( name, w, h, sizemode, format, depth, true, false );

	}

	// push
	ILuaObject* meta = Lua()->GetMetaTable( "ITexture", GLua::TYPE_TEXTURE );
	Lua()->PushUserData( meta, texture );
	SAFE_UNREF( meta );

	return 1;

}

/*-------------------------------------
	GetBackBufferFormat
-------------------------------------*/
LUA_FUNCTION( GetBackBufferFormat )
{

	// create
	Lua()->PushLong( materialsystem->GetBackBufferFormat() );

	return 1;

}

/*-------------------------------------
	ClearStencil
-------------------------------------*/
LUA_FUNCTION( ClearStencil )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->ClearBuffers( false, false, true );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	ClearObeyStencil
-------------------------------------*/
LUA_FUNCTION( ClearObeyStencil )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );

	// arguments
	unsigned char r = static_cast<unsigned char>( Lua()->GetInteger( 1 ) );
	unsigned char g = static_cast<unsigned char>( Lua()->GetInteger( 2 ) );
	unsigned char b = static_cast<unsigned char>( Lua()->GetInteger( 3 ) );
	unsigned char a = static_cast<unsigned char>( Lua()->GetInteger( 4 ) );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->ClearColor4ub( r, g, b, a );
	context->ClearBuffersObeyStencil( true, false );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	ClearDepthObeyStencil
-------------------------------------*/
LUA_FUNCTION( ClearDepthObeyStencil )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->ClearBuffersObeyStencil( false, true );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetFogMode
-------------------------------------*/
LUA_FUNCTION( SetFogMode )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->FogMode( static_cast<MaterialFogMode_t>( Lua()->GetInteger( 1 ) ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetFogStart
-------------------------------------*/
LUA_FUNCTION( SetFogStart )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->FogStart( Lua()->GetNumber( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetFogEnd
-------------------------------------*/
LUA_FUNCTION( SetFogEnd )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->FogEnd( Lua()->GetNumber( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetFogZ
-------------------------------------*/
LUA_FUNCTION( SetFogZ )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetFogZ( Lua()->GetNumber( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	GetFogMode
-------------------------------------*/
LUA_FUNCTION( GetFogMode )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	Lua()->Push( static_cast<float>( context->GetFogMode() ) );
	SAFE_RELEASE( context );

	return 1;

}

/*-------------------------------------
	SetFogColor
-------------------------------------*/
LUA_FUNCTION( SetFogColor )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );

	// arguments
	unsigned char r = static_cast<unsigned char>( Lua()->GetInteger( 1 ) );
	unsigned char g = static_cast<unsigned char>( Lua()->GetInteger( 2 ) );
	unsigned char b = static_cast<unsigned char>( Lua()->GetInteger( 3 ) );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->FogColor3ub( r, g, b );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	GetFogColor
-------------------------------------*/
LUA_FUNCTION( GetFogColor )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();

	// fetch
	unsigned char rgb[ 3 ];
	context->GetFogColor( rgb );

	SAFE_RELEASE( context );

	// push
	Lua()->Push( static_cast<float>( rgb[ 0 ] ) );
	Lua()->Push( static_cast<float>( rgb[ 1 ] ) );
	Lua()->Push( static_cast<float>( rgb[ 2 ] ) );

	return 3;

}


/*-------------------------------------
	GetFogDistances
-------------------------------------*/
LUA_FUNCTION( GetFogDistances )
{

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();

	// fetch
	float start, end, z;
	context->GetFogDistances( &start, &end, &z );

	SAFE_RELEASE( context );

	// push
	Lua()->Push( start );
	Lua()->Push( end );
	Lua()->Push( z );

	return 3;

}

/*-------------------------------------
	SetFogMaxDensity
-------------------------------------*/
LUA_FUNCTION( SetFogMaxDensity )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->FogMaxDensity( Lua()->GetNumber( 1 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	OverrideDepthEnable
-------------------------------------*/
LUA_FUNCTION( OverrideDepthEnable )
{

	Lua()->CheckType( 1, GLua::TYPE_BOOL );
	Lua()->CheckType( 2, GLua::TYPE_BOOL );

	// arguments
	bool enable = Lua()->GetBool( 1 );
	bool depth = Lua()->GetBool( 2 );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->OverrideDepthEnable( enable, depth );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	CullMode
-------------------------------------*/
LUA_FUNCTION( CullMode )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->CullMode( static_cast<MaterialCullMode_t>( Lua()->GetInteger( 1 ) ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	SetScissorRect
-------------------------------------*/
LUA_FUNCTION( SetScissorRect )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );
	Lua()->CheckType( 5, GLua::TYPE_BOOL );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->SetScissorRect( Lua()->GetInteger( 1 ), Lua()->GetInteger( 2 ), Lua()->GetInteger( 3 ), Lua()->GetInteger( 4 ), Lua()->GetBool( 5 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	DepthRange
-------------------------------------*/
LUA_FUNCTION( DepthRange )
{

	Lua()->CheckType( 1, GLua::TYPE_NUMBER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );

	// get the current context
	IMatRenderContext* context = materialsystem->GetRenderContext();
	context->DepthRange( Lua()->GetNumber( 1 ), Lua()->GetNumber( 2 ) );
	SAFE_RELEASE( context );

	return 0;

}

/*-------------------------------------
	GetNumAnimationFrames
-------------------------------------*/
LUA_FUNCTION( GetNumAnimationFrames )
{

	Lua()->CheckType( 1, GLua::TYPE_MATERIAL );

	// arguments
	IMaterial* mat = reinterpret_cast<IMaterial*>( Lua()->GetUserData( 1 ) );
	if( mat )
	{

		// return the number of frames
		Lua()->Push( static_cast<float>( mat->GetNumAnimationFrames() ) );
		return 1;

	}

	return 0;

}

/*-------------------------------------
	IsErrorMaterial
-------------------------------------*/
LUA_FUNCTION( IsErrorMaterial )
{

	Lua()->CheckType( 1, GLua::TYPE_MATERIAL );

	// arguments
	IMaterial* mat = reinterpret_cast<IMaterial*>( Lua()->GetUserData( 1 ) );
	if( mat )
	{

		// error?
		Lua()->Push( mat->IsErrorMaterial() );
		return 1;

	}

	return 0;

}

/*-------------------------------------
	IsErrorTexture
-------------------------------------*/
LUA_FUNCTION( IsErrorTexture )
{

	Lua()->CheckType( 1, GLua::TYPE_TEXTURE );

	// arguments
	ITexture* tex = reinterpret_cast<ITexture*>( Lua()->GetUserData( 1 ) );
	if( tex )
	{

		// error?
		Lua()->Push( tex->IsError() );
		return 1;

	}

	return 0;

}

/*-------------------------------------
	Plane
-------------------------------------*/
LUA_FUNCTION( Plane )
{

	Lua()->CheckType( 1, GLua::TYPE_VECTOR );
	Lua()->CheckType( 2, GLua::TYPE_VECTOR );

	// arguments
	Vector normal = *reinterpret_cast<Vector*>( Lua()->GetUserData( 1 ) );
	Vector point = *reinterpret_cast<Vector*>( Lua()->GetUserData( 2 ) );

	// calculate distance
	float distance = DotProduct( normal, point );

	// return the plane
	ILuaObject* table = Lua()->GetNewTable();

	// set the vector
	ILuaObject* normalobj = Lua()->NewTemporaryObject();
	normalobj->SetFromStack( 1 );
	table->SetMember( "Normal", normalobj );

	SAFE_UNREF( normalobj );

	// set the distance
	table->SetMember( "Distance", distance );

	// push the table
	Lua()->Push( table );

	SAFE_UNREF( table );

	return 1;

}




/*-------------------------------------
	SetRenderClipPlane
-------------------------------------*/
LUA_FUNCTION( SetRenderClipPlane )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get plane struct
	ClipPlane& clipplane = *reinterpret_cast<ClipPlane*>( reinterpret_cast<char*>( entity ) + OFFSET_CLIPPLANE );

	// check type
	switch( Lua()->GetType( 2 ) )
	{

	// vector and distance
	case GLua::TYPE_VECTOR:
		{

			// make sure they passed a distance
			Lua()->CheckType( 3, GLua::TYPE_NUMBER );

			// get vector
			Vector& normal = GetVector( L, 2 );
			float dist = Lua()->GetNumber( 3 );

			// set
			clipplane.plane[ 0 ] = normal.x;
			clipplane.plane[ 1 ] = normal.y;
			clipplane.plane[ 2 ] = normal.z;
			clipplane.plane[ 3 ] = dist;

			break;

		}

	}

	return 0;

}

/*-------------------------------------
	SetRenderClipPlaneEnabled
-------------------------------------*/
LUA_FUNCTION( SetRenderClipPlaneEnabled )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );
	Lua()->CheckType( 2, GLua::TYPE_BOOL );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get plane struct
	ClipPlane& clipplane = *reinterpret_cast<ClipPlane*>( reinterpret_cast<char*>( entity ) + OFFSET_CLIPPLANE );

	// enable
	clipplane.enabled = Lua()->GetBool( 2 );

	return 0;

}


/*-------------------------------------
	GetFogParams
-------------------------------------*/
LUA_FUNCTION( GetFogParams )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get fog params structure
	fogparams_t& fog = *reinterpret_cast<fogparams_t*>( reinterpret_cast<char*>( entity ) + OFFSET_FOGPARAMS );

	ILuaObject* table = Lua()->GetNewTable();

	// primary colors
	ILuaObject* colorPrimary = Lua()->GetNewTable();
		colorPrimary->SetMember( "r", static_cast<float>( fog.colorPrimary.m_Value.r ) );
		colorPrimary->SetMember( "g", static_cast<float>( fog.colorPrimary.m_Value.g ) );
		colorPrimary->SetMember( "b", static_cast<float>( fog.colorPrimary.m_Value.b ) );
		colorPrimary->SetMember( "a", static_cast<float>( fog.colorPrimary.m_Value.a ) );
	table->SetMember( "ColorPrimary", colorPrimary );
	SAFE_UNREF( colorPrimary );

	ILuaObject* colorPrimaryLerpTo = Lua()->GetNewTable();
		colorPrimaryLerpTo->SetMember( "r", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.r ) );
		colorPrimaryLerpTo->SetMember( "g", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.g ) );
		colorPrimaryLerpTo->SetMember( "b", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.b ) );
		colorPrimaryLerpTo->SetMember( "a", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.a ) );
	table->SetMember( "ColorPrimaryLerpTo", colorPrimaryLerpTo );
	SAFE_UNREF( colorPrimaryLerpTo );

	// secondary colors
	ILuaObject* colorSecondary = Lua()->GetNewTable();
		colorSecondary->SetMember( "r", static_cast<float>( fog.colorSecondary.m_Value.r ) );
		colorSecondary->SetMember( "g", static_cast<float>( fog.colorSecondary.m_Value.g ) );
		colorSecondary->SetMember( "b", static_cast<float>( fog.colorSecondary.m_Value.b ) );
		colorSecondary->SetMember( "a", static_cast<float>( fog.colorSecondary.m_Value.a ) );
	table->SetMember( "ColorSecondary", colorSecondary );
	SAFE_UNREF( colorSecondary );

	ILuaObject* colorSecondaryLerpTo = Lua()->GetNewTable();
		colorSecondaryLerpTo->SetMember( "r", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.r ) );
		colorSecondaryLerpTo->SetMember( "g", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.g ) );
		colorSecondaryLerpTo->SetMember( "b", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.b ) );
		colorSecondaryLerpTo->SetMember( "a", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.a ) );
	table->SetMember( "ColorSecondaryLerpTo", colorSecondaryLerpTo );
	SAFE_UNREF( colorSecondaryLerpTo );

	// rest of vars
	table->SetMember( "Direction", NewVectorObject( L, fog.dirPrimary.m_Value ) );
	table->SetMember( "Start", fog.start.m_Value );
	table->SetMember( "StartLerpTo", fog.startLerpTo.m_Value );
	table->SetMember( "End", fog.end.m_Value );
	table->SetMember( "EndLerpTo", fog.endLerpTo.m_Value );
	table->SetMember( "FarZ", fog.farz.m_Value );
	table->SetMember( "MaxDensity", fog.maxdensity.m_Value );
	table->SetMember( "Duration", fog.duration.m_Value );
	table->SetMember( "LerpTime", fog.lerptime.m_Value );
	table->SetMember( "Enable", fog.enable.m_Value );
	table->SetMember( "Blend", fog.blend.m_Value );

	Lua()->Push( table );

	SAFE_UNREF( table );

	return 1;

}

/*-------------------------------------
	SetFogParams
-------------------------------------*/
LUA_FUNCTION( SetFogParams )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );
	Lua()->CheckType( 2, GLua::TYPE_TABLE );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get fog params structure
	fogparams_t& fog = *reinterpret_cast<fogparams_t*>( reinterpret_cast<char*>( entity ) + OFFSET_FOGPARAMS );

	ILuaObject* table = Lua()->GetObject( 2 );

	// apply their settings
	fog.blend.m_Value = table->GetMemberBool( "Blend", false );
	fog.enable.m_Value = table->GetMemberBool( "Enable", false );
	fog.lerptime.m_Value = table->GetMemberFloat( "LerpTime", 0 );
	fog.duration.m_Value = table->GetMemberFloat( "Duration", 0 );
	fog.maxdensity.m_Value = table->GetMemberFloat( "MaxDensity", 1 );
	fog.farz.m_Value = table->GetMemberFloat( "FarZ", -1 );
	fog.start.m_Value = table->GetMemberFloat( "Start", 100 );
	fog.startLerpTo.m_Value = table->GetMemberFloat( "StartLerpTo", 100 );
	fog.end.m_Value = table->GetMemberFloat( "End", 10000 );
	fog.endLerpTo.m_Value = table->GetMemberFloat( "EndLerpTo", 10000 );

	// Direction
	ILuaObject* temp = table->GetMember( "Direction" );
	if( temp && !temp->isNil() )
		fog.dirPrimary.m_Value = *reinterpret_cast<Vector*>( temp->GetUserData() );
	SAFE_UNREF( temp );

	// ColorPrimary
	temp = table->GetMember( "ColorPrimary" );
	if( temp && !temp->isNil() )
	{

		fog.colorPrimary.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorPrimary.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorPrimary.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorPrimaryLerpTo
	temp = table->GetMember( "ColorPrimaryLerpTo" );
	if( temp && !temp->isNil() )
	{

		fog.colorPrimaryLerpTo.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorPrimaryLerpTo.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorPrimaryLerpTo.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorSecondary
	temp = table->GetMember( "ColorSecondary" );
	if( temp && !temp->isNil() )
	{

		fog.colorSecondary.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorSecondary.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorSecondary.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorSecondaryLerpTo
	temp = table->GetMember( "ColorSecondaryLerpTo" );
	if( temp && !temp->isNil() )
	{

		fog.colorSecondaryLerpTo.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorSecondaryLerpTo.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorSecondaryLerpTo.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	SAFE_UNREF( table );

	return 0;

}

/*-------------------------------------
	GetFogParamsSkybox
-------------------------------------*/
LUA_FUNCTION( GetFogParamsSkybox )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get fog params structure
	fogparams_t& fog = (*reinterpret_cast<CPlayerLocalData*>( reinterpret_cast<char*>( entity ) + OFFSET_LOCALDATA )).m_skybox3d.fog;

	ILuaObject* table = Lua()->GetNewTable();

	// primary colors
	ILuaObject* colorPrimary = Lua()->GetNewTable();
		colorPrimary->SetMember( "r", static_cast<float>( fog.colorPrimary.m_Value.r ) );
		colorPrimary->SetMember( "g", static_cast<float>( fog.colorPrimary.m_Value.g ) );
		colorPrimary->SetMember( "b", static_cast<float>( fog.colorPrimary.m_Value.b ) );
		colorPrimary->SetMember( "a", static_cast<float>( fog.colorPrimary.m_Value.a ) );
	table->SetMember( "ColorPrimary", colorPrimary );
	SAFE_UNREF( colorPrimary );

	ILuaObject* colorPrimaryLerpTo = Lua()->GetNewTable();
		colorPrimaryLerpTo->SetMember( "r", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.r ) );
		colorPrimaryLerpTo->SetMember( "g", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.g ) );
		colorPrimaryLerpTo->SetMember( "b", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.b ) );
		colorPrimaryLerpTo->SetMember( "a", static_cast<float>( fog.colorPrimaryLerpTo.m_Value.a ) );
	table->SetMember( "ColorPrimaryLerpTo", colorPrimaryLerpTo );
	SAFE_UNREF( colorPrimaryLerpTo );

	// secondary colors
	ILuaObject* colorSecondary = Lua()->GetNewTable();
		colorSecondary->SetMember( "r", static_cast<float>( fog.colorSecondary.m_Value.r ) );
		colorSecondary->SetMember( "g", static_cast<float>( fog.colorSecondary.m_Value.g ) );
		colorSecondary->SetMember( "b", static_cast<float>( fog.colorSecondary.m_Value.b ) );
		colorSecondary->SetMember( "a", static_cast<float>( fog.colorSecondary.m_Value.a ) );
	table->SetMember( "ColorSecondary", colorSecondary );
	SAFE_UNREF( colorSecondary );

	ILuaObject* colorSecondaryLerpTo = Lua()->GetNewTable();
		colorSecondaryLerpTo->SetMember( "r", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.r ) );
		colorSecondaryLerpTo->SetMember( "g", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.g ) );
		colorSecondaryLerpTo->SetMember( "b", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.b ) );
		colorSecondaryLerpTo->SetMember( "a", static_cast<float>( fog.colorSecondaryLerpTo.m_Value.a ) );
	table->SetMember( "ColorSecondaryLerpTo", colorSecondaryLerpTo );
	SAFE_UNREF( colorSecondaryLerpTo );

	// rest of vars
	table->SetMember( "Direction", NewVectorObject( L, fog.dirPrimary.m_Value ) );
	table->SetMember( "Start", fog.start.m_Value );
	table->SetMember( "StartLerpTo", fog.startLerpTo.m_Value );
	table->SetMember( "End", fog.end.m_Value );
	table->SetMember( "EndLerpTo", fog.endLerpTo.m_Value );
	table->SetMember( "FarZ", fog.farz.m_Value );
	table->SetMember( "MaxDensity", fog.maxdensity.m_Value );
	table->SetMember( "Duration", fog.duration.m_Value );
	table->SetMember( "LerpTime", fog.lerptime.m_Value );
	table->SetMember( "Enable", fog.enable.m_Value );
	table->SetMember( "Blend", fog.blend.m_Value );

	Lua()->Push( table );

	SAFE_UNREF( table );

	return 1;

}

/*-------------------------------------
	SetFogParamsSkybox
-------------------------------------*/
LUA_FUNCTION( SetFogParamsSkybox )
{

	Lua()->CheckType( 1, GLua::TYPE_ENTITY );
	Lua()->CheckType( 2, GLua::TYPE_TABLE );

	// get entity
	C_BaseEntity* entity = reinterpret_cast<C_BaseEntity*>( cliententitylist->GetClientEntityFromHandle( *reinterpret_cast<CBaseHandle*>( Lua()->GetUserData( 1 ) ) ) );

	// get fog params structure
	fogparams_t& fog = (*reinterpret_cast<CPlayerLocalData*>( reinterpret_cast<char*>( entity ) + OFFSET_LOCALDATA )).m_skybox3d.fog;

	ILuaObject* table = Lua()->GetObject( 2 );

	// apply their settings
	fog.blend.m_Value = table->GetMemberBool( "Blend", false );
	fog.enable.m_Value = table->GetMemberBool( "Enable", false );
	fog.lerptime.m_Value = table->GetMemberFloat( "LerpTime", 0 );
	fog.duration.m_Value = table->GetMemberFloat( "Duration", 0 );
	fog.maxdensity.m_Value = table->GetMemberFloat( "MaxDensity", 1 );
	fog.farz.m_Value = table->GetMemberFloat( "FarZ", -1 );
	fog.start.m_Value = table->GetMemberFloat( "Start", 100 );
	fog.startLerpTo.m_Value = table->GetMemberFloat( "StartLerpTo", 100 );
	fog.end.m_Value = table->GetMemberFloat( "End", 10000 );
	fog.endLerpTo.m_Value = table->GetMemberFloat( "EndLerpTo", 10000 );

	// Direction
	ILuaObject* temp = table->GetMember( "Direction" );
	if( temp && !temp->isNil() )
		fog.dirPrimary.m_Value = *reinterpret_cast<Vector*>( temp->GetUserData() );
	SAFE_UNREF( temp );

	// ColorPrimary
	temp = table->GetMember( "ColorPrimary" );
	if( temp && !temp->isNil() )
	{

		fog.colorPrimary.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorPrimary.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorPrimary.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorPrimaryLerpTo
	temp = table->GetMember( "ColorPrimaryLerpTo" );
	if( temp && !temp->isNil() )
	{

		fog.colorPrimaryLerpTo.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorPrimaryLerpTo.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorPrimaryLerpTo.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorSecondary
	temp = table->GetMember( "ColorSecondary" );
	if( temp && !temp->isNil() )
	{

		fog.colorSecondary.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorSecondary.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorSecondary.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	// ColorSecondaryLerpTo
	temp = table->GetMember( "ColorSecondaryLerpTo" );
	if( temp && !temp->isNil() )
	{

		fog.colorSecondaryLerpTo.m_Value.r = temp->GetMemberInt( "r", 255 );
		fog.colorSecondaryLerpTo.m_Value.g = temp->GetMemberInt( "g", 255 );
		fog.colorSecondaryLerpTo.m_Value.b = temp->GetMemberInt( "b", 255 );

	}
	SAFE_UNREF( temp );

	SAFE_UNREF( table );

	return 0;

}

/*-------------------------------------
	GetDynamicMesh
-------------------------------------*/
LUA_FUNCTION( GetDynamicMesh )
{

	Lua()->CheckType( 1, GLua::TYPE_BOOL );

	IMatRenderContext* context = materialsystem->GetRenderContext();

		IMesh* mesh = context->GetDynamicMesh( Lua()->GetBool( 1 ) );

		// push
		ILuaObject* meta = Lua()->GetMetaTable( "Mesh", TYPE_MESH );
		Lua()->PushUserData( meta, mesh );
		SAFE_UNREF( meta );

	SAFE_RELEASE( context );

	return 1;

}


/*-------------------------------------
	MeshBuilderBegin
-------------------------------------*/
LUA_FUNCTION( MeshBuilderBegin )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, TYPE_MESH );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->Begin( reinterpret_cast<IMesh*>( Lua()->GetUserData( 2 ) ), static_cast<MaterialPrimitiveType_t>( Lua()->GetInteger( 3 ) ), Lua()->GetInteger( 4 ) );

	return 0;

}


/*-------------------------------------
	MeshBuilderEnd
-------------------------------------*/
LUA_FUNCTION( MeshBuilderEnd )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->End( false, false );

	return 0;

}


/*-------------------------------------
	MeshBuilderGC
-------------------------------------*/
LUA_FUNCTION( MeshBuilderGC )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );
	delete builder;

	return 0;

}


/*-------------------------------------
	MeshBuilderAdvanceVertex
-------------------------------------*/
LUA_FUNCTION( MeshBuilderAdvanceVertex )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->AdvanceVertex();

	return 0;

}

/*-------------------------------------
	MeshBuilderPosition
-------------------------------------*/
LUA_FUNCTION( MeshBuilderPosition )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_VECTOR );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->Position3fv( reinterpret_cast<float*>( Lua()->GetUserData( 2 ) ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderNormal
-------------------------------------*/
LUA_FUNCTION( MeshBuilderNormal )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_VECTOR );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->Normal3fv( reinterpret_cast<float*>( Lua()->GetUserData( 2 ) ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderColor
-------------------------------------*/
LUA_FUNCTION( MeshBuilderColor )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_TABLE );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	ILuaObject* color = Lua()->GetObject( 2 );
	builder->Color4ub( color->GetMemberInt( "r", 255 ), color->GetMemberInt( "g", 255 ), color->GetMemberInt( "b", 255 ), color->GetMemberInt( "a", 255 ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderTexCoord
-------------------------------------*/
LUA_FUNCTION( MeshBuilderTexCoord )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_NUMBER );
	Lua()->CheckType( 3, GLua::TYPE_NUMBER );
	Lua()->CheckType( 4, GLua::TYPE_NUMBER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->TexCoord2f( Lua()->GetInteger( 2 ), Lua()->GetNumber( 3 ), Lua()->GetNumber( 4 ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderSpecular
-------------------------------------*/
LUA_FUNCTION( MeshBuilderSpecular )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_TABLE );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	ILuaObject* color = Lua()->GetObject( 2 );
	builder->Specular4ub( color->GetMemberInt( "r", 255 ), color->GetMemberInt( "g", 255 ), color->GetMemberInt( "b", 255 ), color->GetMemberInt( "a", 255 ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderTangentS
-------------------------------------*/
LUA_FUNCTION( MeshBuilderTangentS )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_VECTOR );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->TangentS3fv( reinterpret_cast<float*>( Lua()->GetUserData( 2 ) ) );

	return 0;

}


/*-------------------------------------
	MeshBuilderTangentT
-------------------------------------*/
LUA_FUNCTION( MeshBuilderTangentT )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );
	Lua()->CheckType( 2, GLua::TYPE_VECTOR );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->TangentT3fv( reinterpret_cast<float*>( Lua()->GetUserData( 2 ) ) );

	return 0;

}

/*-------------------------------------
	MeshBuilderReset
-------------------------------------*/
LUA_FUNCTION( MeshBuilderReset )
{

	Lua()->CheckType( 1, TYPE_MESHBUILDER );

	CMeshBuilder* builder = reinterpret_cast<CMeshBuilder*>( Lua()->GetUserData( 1 ) );

	builder->Reset();

	return 0;

}


/*-------------------------------------
	MeshBuilderCreate
-------------------------------------*/
LUA_FUNCTION( MeshBuilderCreate )
{

	CMeshBuilder* builder = new CMeshBuilder();

	// push
	ILuaObject* meta = Lua()->GetMetaTable( "CMeshBuilder", TYPE_MESHBUILDER );
	Lua()->PushUserData( meta, builder );
	SAFE_UNREF( meta );

	return 1;

}

/*-------------------------------------
	MeshDraw
-------------------------------------*/
LUA_FUNCTION( MeshDraw )
{

	Lua()->CheckType( 1, TYPE_MESH );

	IMesh* mesh = reinterpret_cast<IMesh*>( Lua()->GetUserData( 1 ) );

	mesh->Draw();

	return 0;

}


/*-------------------------------------
	Init
-------------------------------------*/
int Init( lua_State* L )
{

	CreateInterfaceFn EngineFactory = Sys_GetFactory( "MaterialSystem.dll" );
	CreateInterfaceFn ClientFactory = Sys_GetFactory( "client.dll" );

	// get the server interface
	materialsystem = (IMaterialSystem*)EngineFactory( MATERIAL_SYSTEM_INTERFACE_VERSION, NULL );
	if( !materialsystem )
		Lua()->Error( "Failed to get required interfaces.\n" );
	
	// get the entity list interface
	cliententitylist = (IClientEntityList*)ClientFactory( VCLIENTENTITYLIST_INTERFACE_VERSION, NULL );
	if( !cliententitylist )
		Lua()->Error( "Failed to get required interfaces.\n" );

	ILuaObject* metatable;
	ILuaObject* __index;

	// additional render methods
	ILuaObject* render = Lua()->GetGlobal( "render" );
		render->SetMember( "OverrideDepthEnable", OverrideDepthEnable );
		render->SetMember( "PushCustomClipPlane", PushCustomClipPlane );
		render->SetMember( "PopCustomClipPlane", PopCustomClipPlane );
		render->SetMember( "EnableClipping", EnableClipping );
		render->SetMember( "CreateRenderTargetTexture", CreateRenderTargetTexture );
		render->SetMember( "GetBackBufferFormat", GetBackBufferFormat );
		render->SetMember( "SetStencilEnable", SetStencilEnable );
		render->SetMember( "SetStencilFailOperation", SetStencilFailOperation );
		render->SetMember( "SetStencilZFailOperation", SetStencilZFailOperation );
		render->SetMember( "SetStencilPassOperation", SetStencilPassOperation );
		render->SetMember( "SetStencilCompareFunction", SetStencilCompareFunction );
		render->SetMember( "SetStencilReferenceValue", SetStencilReferenceValue );
		render->SetMember( "SetStencilTestMask", SetStencilTestMask );
		render->SetMember( "SetStencilWriteMask", SetStencilWriteMask );
		render->SetMember( "ClearStencilBufferRectangle", ClearStencilBufferRectangle );
		render->SetMember( "ClearStencil", ClearStencil );
		render->SetMember( "ClearObeyStencil", ClearObeyStencil );
		render->SetMember( "ClearDepthObeyStencil", ClearDepthObeyStencil );
		render->SetMember( "GetFogDistances", GetFogDistances );
		render->SetMember( "GetFogColor", GetFogColor );
		render->SetMember( "SetFogColor", SetFogColor );
		render->SetMember( "GetFogMode", GetFogMode );
		render->SetMember( "SetFogZ", SetFogZ );
		render->SetMember( "SetFogEnd", SetFogEnd );
		render->SetMember( "SetFogStart", SetFogStart );
		render->SetMember( "SetFogMaxDensity", SetFogMaxDensity );
		render->SetMember( "SetFogMode", SetFogMode );
		render->SetMember( "CullMode", CullMode );
		render->SetMember( "SetScissorRect", SetScissorRect );
		render->SetMember( "DepthRange", DepthRange );
		render->SetMember( "GetDynamicMesh", GetDynamicMesh );
	render->UnReference();

	// extend IMaterial metatable
	metatable = Lua()->GetMetaTable( "IMaterial", GLua::TYPE_MATERIAL );
		metatable->SetMember( "GetNumAnimationFrames", GetNumAnimationFrames );
		metatable->SetMember( "IsError", IsErrorMaterial );
	metatable->UnReference();

	// extend ITexture metatable
	metatable = Lua()->GetMetaTable( "ITexture", GLua::TYPE_TEXTURE );
		metatable->SetMember( "IsError", IsErrorMaterial );
	metatable->UnReference();

	// extend Entity metatable
	metatable = Lua()->GetMetaTable( "Entity", GLua::TYPE_ENTITY );
		metatable->SetMember( "SetRenderClipPlane", SetRenderClipPlane );
		metatable->SetMember( "SetRenderClipPlaneEnabled", SetRenderClipPlaneEnabled );
	metatable->UnReference();

	// extend Player metatable
	metatable = Lua()->GetMetaTable( "Player", GLua::TYPE_ENTITY );
		metatable->SetMember( "GetFogParams", GetFogParams );
		metatable->SetMember( "SetFogParams", SetFogParams );
		metatable->SetMember( "GetFogParamsSkybox", GetFogParamsSkybox );
		metatable->SetMember( "SetFogParamsSkybox", SetFogParamsSkybox );
	metatable->UnReference();


	// mesh builder
	metatable = Lua()->GetMetaTable( "CMeshBuilder", TYPE_MESHBUILDER );
		__index = Lua()->GetNewTable();
			__index->SetMember( "Begin", MeshBuilderBegin );
			__index->SetMember( "End", MeshBuilderEnd );
			__index->SetMember( "AdvanceVertex", MeshBuilderAdvanceVertex );
			__index->SetMember( "Position", MeshBuilderPosition );
			__index->SetMember( "Normal", MeshBuilderNormal );
			__index->SetMember( "Color", MeshBuilderColor );
			__index->SetMember( "TexCoord", MeshBuilderTexCoord );
			__index->SetMember( "Specular", MeshBuilderSpecular );
			__index->SetMember( "TangentS", MeshBuilderTangentS );
			__index->SetMember( "TangentT", MeshBuilderTangentT );
			__index->SetMember( "Reset", MeshBuilderReset );
		metatable->SetMember( "__index", __index );
		__index->UnReference();
		metatable->SetMember( "__gc", MeshBuilderGC );
	metatable->UnReference();

	// dynamic mesh
	metatable = Lua()->GetMetaTable( "Mesh", TYPE_MESH );
		__index = Lua()->GetNewTable();
			__index->SetMember( "Draw", MeshDraw );
		metatable->SetMember( "__index", __index );
		__index->UnReference();
	metatable->UnReference();

	// global functions
	Lua()->SetGlobal( "MeshBuilder", MeshBuilderCreate );


	// material primitives
	Lua()->SetGlobal( "MATERIAL_POINTS", static_cast<float>( MATERIAL_POINTS ) );
	Lua()->SetGlobal( "MATERIAL_LINES", static_cast<float>( MATERIAL_LINES ) );
	Lua()->SetGlobal( "MATERIAL_TRIANGLES", static_cast<float>( MATERIAL_TRIANGLES ) );
	Lua()->SetGlobal( "MATERIAL_TRIANGLE_STRIP", static_cast<float>( MATERIAL_TRIANGLE_STRIP ) );
	Lua()->SetGlobal( "MATERIAL_LINE_STRIP", static_cast<float>( MATERIAL_LINE_STRIP ) );
	Lua()->SetGlobal( "MATERIAL_LINE_LOOP", static_cast<float>( MATERIAL_LINE_LOOP ) );
	Lua()->SetGlobal( "MATERIAL_POLYGON", static_cast<float>( MATERIAL_POLYGON ) );
	Lua()->SetGlobal( "MATERIAL_QUADS", static_cast<float>( MATERIAL_QUADS ) );
	Lua()->SetGlobal( "MATERIAL_INSTANCED_QUADS", static_cast<float>( MATERIAL_INSTANCED_QUADS ) );
	Lua()->SetGlobal( "MATERIAL_HETEROGENOUS", static_cast<float>( MATERIAL_HETEROGENOUS ) );

	// stencil operation globals
	Lua()->SetGlobal( "STENCILOPERATION_KEEP", static_cast<float>( STENCILOPERATION_KEEP ) );
	Lua()->SetGlobal( "STENCILOPERATION_ZERO", static_cast<float>( STENCILOPERATION_ZERO ) );
	Lua()->SetGlobal( "STENCILOPERATION_REPLACE", static_cast<float>( STENCILOPERATION_REPLACE ) );
	Lua()->SetGlobal( "STENCILOPERATION_INCRSAT", static_cast<float>( STENCILOPERATION_INCRSAT ) );
	Lua()->SetGlobal( "STENCILOPERATION_DECRSAT", static_cast<float>( STENCILOPERATION_DECRSAT ) );
	Lua()->SetGlobal( "STENCILOPERATION_INVERT", static_cast<float>( STENCILOPERATION_INVERT ) );
	Lua()->SetGlobal( "STENCILOPERATION_INCR", static_cast<float>( STENCILOPERATION_INCR ) );
	Lua()->SetGlobal( "STENCILOPERATION_DECR", static_cast<float>( STENCILOPERATION_DECR ) );
	Lua()->SetGlobal( "STENCILOPERATION_FORCE_DWORD", static_cast<float>( STENCILOPERATION_FORCE_DWORD ) );

	// stincel compar functions
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_NEVER", static_cast<float>( STENCILCOMPARISONFUNCTION_NEVER ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_LESS", static_cast<float>( STENCILCOMPARISONFUNCTION_LESS ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_EQUAL", static_cast<float>( STENCILCOMPARISONFUNCTION_EQUAL ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_LESSEQUAL", static_cast<float>( STENCILCOMPARISONFUNCTION_LESSEQUAL ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_GREATER", static_cast<float>( STENCILCOMPARISONFUNCTION_GREATER ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_NOTEQUAL", static_cast<float>( STENCILCOMPARISONFUNCTION_NOTEQUAL ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_GREATEREQUAL", static_cast<float>( STENCILCOMPARISONFUNCTION_GREATEREQUAL ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_ALWAYS", static_cast<float>( STENCILCOMPARISONFUNCTION_ALWAYS ) );
	Lua()->SetGlobal( "STENCILCOMPARISONFUNCTION_FORCE_DWORD", static_cast<float>( STENCILCOMPARISONFUNCTION_FORCE_DWORD ) );

	// size mode
	Lua()->SetGlobal( "RT_SIZE_NO_CHANGE", static_cast<float>( RT_SIZE_NO_CHANGE ) );
	Lua()->SetGlobal( "RT_SIZE_DEFAULT", static_cast<float>( RT_SIZE_DEFAULT ) );
	Lua()->SetGlobal( "RT_SIZE_PICMIP", static_cast<float>( RT_SIZE_PICMIP ) );
	Lua()->SetGlobal( "RT_SIZE_HDR", static_cast<float>( RT_SIZE_HDR ) );
	Lua()->SetGlobal( "RT_SIZE_FULL_FRAME_BUFFER", static_cast<float>( RT_SIZE_FULL_FRAME_BUFFER ) );
	Lua()->SetGlobal( "RT_SIZE_OFFSCREEN", static_cast<float>( RT_SIZE_OFFSCREEN ) );
	Lua()->SetGlobal( "RT_SIZE_FULL_FRAME_BUFFER_ROUNDED_UP", static_cast<float>( RT_SIZE_FULL_FRAME_BUFFER_ROUNDED_UP ) );

	// image format
	Lua()->SetGlobal( "IMAGE_FORMAT_UNKNOWN", static_cast<float>( IMAGE_FORMAT_UNKNOWN ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGBA8888", static_cast<float>( IMAGE_FORMAT_RGBA8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ABGR8888", static_cast<float>( IMAGE_FORMAT_ABGR8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGB888", static_cast<float>( IMAGE_FORMAT_RGB888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGR888", static_cast<float>( IMAGE_FORMAT_BGR888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGB565", static_cast<float>( IMAGE_FORMAT_RGB565 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_I8", static_cast<float>( IMAGE_FORMAT_I8 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_IA88", static_cast<float>( IMAGE_FORMAT_IA88 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_P8", static_cast<float>( IMAGE_FORMAT_P8 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_A8", static_cast<float>( IMAGE_FORMAT_A8 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGB888_BLUESCREEN", static_cast<float>( IMAGE_FORMAT_RGB888_BLUESCREEN ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGR888_BLUESCREEN", static_cast<float>( IMAGE_FORMAT_BGR888_BLUESCREEN ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ARGB8888", static_cast<float>( IMAGE_FORMAT_ARGB8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGRA8888", static_cast<float>( IMAGE_FORMAT_BGRA8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_DXT1", static_cast<float>( IMAGE_FORMAT_DXT1 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_DXT3", static_cast<float>( IMAGE_FORMAT_DXT3 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_DXT5", static_cast<float>( IMAGE_FORMAT_DXT5 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGRX8888", static_cast<float>( IMAGE_FORMAT_BGRX8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGR565", static_cast<float>( IMAGE_FORMAT_BGR565 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGRX5551", static_cast<float>( IMAGE_FORMAT_BGRX5551 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGRA4444", static_cast<float>( IMAGE_FORMAT_BGRA4444 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_DXT1_ONEBITALPHA", static_cast<float>( IMAGE_FORMAT_DXT1_ONEBITALPHA ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_BGRA5551", static_cast<float>( IMAGE_FORMAT_BGRA5551 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_UV88", static_cast<float>( IMAGE_FORMAT_UV88 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_UVWQ8888", static_cast<float>( IMAGE_FORMAT_UVWQ8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGBA16161616F", static_cast<float>( IMAGE_FORMAT_RGBA16161616F ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGBA16161616", static_cast<float>( IMAGE_FORMAT_RGBA16161616 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_UVLX8888", static_cast<float>( IMAGE_FORMAT_UVLX8888 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_R32F", static_cast<float>( IMAGE_FORMAT_R32F ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGB323232F", static_cast<float>( IMAGE_FORMAT_RGB323232F ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_RGBA32323232F", static_cast<float>( IMAGE_FORMAT_RGBA32323232F ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_NV_DST16", static_cast<float>( IMAGE_FORMAT_NV_DST16 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_NV_DST24", static_cast<float>( IMAGE_FORMAT_NV_DST24 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_NV_INTZ", static_cast<float>( IMAGE_FORMAT_NV_INTZ ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_NV_RAWZ", static_cast<float>( IMAGE_FORMAT_NV_RAWZ ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ATI_DST16", static_cast<float>( IMAGE_FORMAT_ATI_DST16 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ATI_DST24", static_cast<float>( IMAGE_FORMAT_ATI_DST24 ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_NV_NULL", static_cast<float>( IMAGE_FORMAT_NV_NULL ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ATI2N", static_cast<float>( IMAGE_FORMAT_ATI2N ) );
	Lua()->SetGlobal( "IMAGE_FORMAT_ATI1N", static_cast<float>( IMAGE_FORMAT_ATI1N ) );

	// material fog mode
	Lua()->SetGlobal( "MATERIAL_FOG_NONE", static_cast<float>( MATERIAL_FOG_NONE ) );
	Lua()->SetGlobal( "MATERIAL_FOG_LINEAR", static_cast<float>( MATERIAL_FOG_LINEAR ) );
	Lua()->SetGlobal( "MATERIAL_FOG_LINEAR_BELOW_FOG_Z", static_cast<float>( MATERIAL_FOG_LINEAR_BELOW_FOG_Z ) );

	// material cull mode
	Lua()->SetGlobal( "MATERIAL_CULLMODE_CCW", static_cast<float>( MATERIAL_CULLMODE_CCW ) );
	Lua()->SetGlobal( "MATERIAL_CULLMODE_CW", static_cast<float>( MATERIAL_CULLMODE_CW ) );

	// material rt depth
	Lua()->SetGlobal( "MATERIAL_RT_DEPTH_SHARED", static_cast<float>( MATERIAL_RT_DEPTH_SHARED ) );
	Lua()->SetGlobal( "MATERIAL_RT_DEPTH_SEPARATE", static_cast<float>( MATERIAL_RT_DEPTH_SEPARATE ) );
	Lua()->SetGlobal( "MATERIAL_RT_DEPTH_NONE", static_cast<float>( MATERIAL_RT_DEPTH_NONE ) );
	Lua()->SetGlobal( "MATERIAL_RT_DEPTH_ONLY", static_cast<float>( MATERIAL_RT_DEPTH_ONLY ) );

	return 0;

}

/*-------------------------------------
	Shutdown
-------------------------------------*/
int Shutdown( lua_State* L )
{

	return 0;

}
